@@ -31,22 +31,22 @@
 #include <time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/socket.h>
-#include <sys/un.h>
 #include <fcntl.h>
 #include <inttypes.h>
 #ifdef HAVE_GETOPT_LONG 
 #include <getopt.h>
 #endif
-
 #ifdef HAVE_LIBCAP_NG
 #include <cap-ng.h>
 #endif
+#ifdef HAVE_IRQBALANCEUI
+#include <sys/un.h>
+#include <sys/socket.h>
+#endif
 #include "irqbalance.h"
+#include "thermal.h"
 
 volatile int keep_going = 1;
-int socket_fd;
-char socket_name[64];
 int one_shot_mode;
 int debug_mode;
 int foreground_mode;
@@ -67,9 +67,14 @@ int last_interval;
 GMainLoop *main_loop;
 
 char *cpu_ban_string = NULL;
-char *banned_cpumask_from_ui = NULL;
 unsigned long migrate_ratio = 0;
 
+#ifdef HAVE_IRQBALANCEUI
+int socket_fd;
+char socket_name[64];
+char *banned_cpumask_from_ui = NULL;
+#endif
+
 static void sleep_approx(int seconds)
 {
 	struct timespec ts;
@@ -119,6 +124,7 @@ static void parse_command_line(int argc, char **argv)
 	int opt;
 	int longind;
 	unsigned long val;
+	char *endptr;
 
 	while ((opt = getopt_long(argc, argv,
 		"odfjVi:p:s:c:l:m:t:e:",
@@ -134,8 +140,8 @@ static void parse_command_line(int argc, char **argv)
 				exit(1);
 				break;
 			case 'c':
-				deepest_cache = strtoul(optarg, NULL, 10);
-				if (deepest_cache == ULONG_MAX || deepest_cache < 1) {
+				deepest_cache = strtoul(optarg, &endptr, 10);
+				if (optarg == endptr || deepest_cache == ULONG_MAX || deepest_cache < 1) {
 					usage();
 					exit(1);
 				}
@@ -148,8 +154,8 @@ static void parse_command_line(int argc, char **argv)
 				foreground_mode=1;
 				break;
 			case 'i':
-				val = strtoull(optarg, NULL, 10);
-				if (val == ULONG_MAX) {
+				val = strtoull(optarg, &endptr, 10);
+				if (optarg == endptr || val == ULONG_MAX) {
 					usage();
 					exit(1);
 				}
@@ -166,8 +172,8 @@ static void parse_command_line(int argc, char **argv)
 				if (!strncmp(optarg, "off", strlen(optarg)))
 					power_thresh = ULONG_MAX;
 				else {
-					power_thresh = strtoull(optarg, NULL, 10);
-					if (power_thresh == ULONG_MAX) {
+					power_thresh = strtoull(optarg, &endptr, 10);
+					if (optarg == endptr || power_thresh == ULONG_MAX) {
 						usage();
 						exit(1);
 					}
@@ -184,14 +190,18 @@ static void parse_command_line(int argc, char **argv)
 				foreground_mode=1;
 				break;
 			case 't':
-				sleep_interval = strtol(optarg, NULL, 10);
-				if (sleep_interval < 1) {
+				sleep_interval = strtol(optarg, &endptr, 10);
+				if (optarg == endptr || sleep_interval < 1) {
 					usage();
 					exit(1);
 				}
 				break;
 			case 'e':
-				migrate_ratio = strtoul(optarg, NULL, 10);
+				migrate_ratio = strtoul(optarg, &endptr, 10);
+				if (optarg == endptr) {
+					usage();
+					exit(1);
+				}
 				break;
 		}
 	}
@@ -285,7 +295,7 @@ gboolean scan(gpointer data __attribute__((unused)))
 
 
 	/* cope with cpu hotplug -- detected during /proc/interrupts parsing */
-	while (need_rescan || need_rebuild) {
+	while (keep_going && (need_rescan || need_rebuild)) {
 		int try_times = 0;
 
 		need_rescan = 0;
@@ -405,6 +415,7 @@ void get_object_stat(struct topo_obj *object, void *data)
 	}
 }
 
+#ifdef HAVE_IRQBALANCEUI
 gboolean sock_handle(gint fd, GIOCondition condition, gpointer user_data __attribute__((unused)))
 {
 	char buff[500];
@@ -458,12 +469,13 @@ gboolean sock_handle(gint fd, GIOCondition condition, gpointer user_data __attri
 			if (!(strncmp(buff + strlen("settings "), "sleep ",
 							strlen("sleep ")))) {
 				char *sleep_string = malloc(
-						sizeof(char) * (recv_size - strlen("settings sleep ")));
+						sizeof(char) * (recv_size - strlen("settings sleep ") + 1));
 
 				if (!sleep_string)
 					goto out_close;
 				strncpy(sleep_string, buff + strlen("settings sleep "),
 						recv_size - strlen("settings sleep "));
+				sleep_string[recv_size - strlen("settings sleep ")] = '\0';
 				int new_iterval = strtoul(sleep_string, NULL, 10);
 				if (new_iterval >= 1) {
 					sleep_interval = new_iterval;
@@ -473,12 +485,13 @@ gboolean sock_handle(gint fd, GIOCondition condition, gpointer user_data __attri
 							strlen("ban irqs ")))) {
 				char *end;
 				char *irq_string = malloc(
-						sizeof(char) * (recv_size - strlen("settings ban irqs ")));
+						sizeof(char) * (recv_size - strlen("settings ban irqs ") + 1));
 
 				if (!irq_string)
 					goto out_close;
 				strncpy(irq_string, buff + strlen("settings ban irqs "),
 						recv_size - strlen("settings ban irqs "));
+				irq_string[recv_size - strlen("settings ban irqs ")] = '\0';
 				g_list_free_full(cl_banned_irqs, free);
 				cl_banned_irqs = NULL;
 				need_rescan = 1;
@@ -498,12 +511,13 @@ gboolean sock_handle(gint fd, GIOCondition condition, gpointer user_data __attri
 				cpu_ban_string = NULL;
 
 				cpu_ban_string = malloc(
-						sizeof(char) * (recv_size - strlen("settings cpus ")));
+						sizeof(char) * (recv_size - strlen("settings cpus ") + 1));
 
 				if (!cpu_ban_string)
 					goto out_close;
 				strncpy(cpu_ban_string, buff + strlen("settings cpus "),
 						recv_size - strlen("settings cpus "));
+				cpu_ban_string[recv_size - strlen("settings cpus ")] = '\0';
 				banned_cpumask_from_ui = strtok(cpu_ban_string, " ");
 				if (!strncmp(banned_cpumask_from_ui, "NULL", strlen("NULL"))) {
 					banned_cpumask_from_ui = NULL;
@@ -585,6 +599,7 @@ int init_socket()
 	g_unix_fd_add(socket_fd, G_IO_IN, sock_handle, NULL);
 	return 0;
 }
+#endif
 
 int main(int argc, char** argv)
 {
@@ -688,10 +703,14 @@ int main(int argc, char** argv)
 	parse_proc_interrupts();
 	parse_proc_stat();
 
+#ifdef HAVE_IRQBALANCEUI
 	if (init_socket()) {
 		ret = EXIT_FAILURE;
 		goto out;
 	}
+#endif
+	if (init_thermal())
+		log(TO_ALL, LOG_WARNING, "Failed to initialize thermal events.\n");
 	main_loop = g_main_loop_new(NULL, FALSE);
 	last_interval = sleep_interval;
 	g_timeout_add_seconds(sleep_interval, scan, NULL);
@@ -700,6 +719,7 @@ int main(int argc, char** argv)
 	g_main_loop_quit(main_loop);
 
 out:
+	deinit_thermal();
 	free_object_tree();
 	free_cl_opts();
 	free(polscript);
@@ -707,11 +727,12 @@ out:
 	/* Remove pidfile */
 	if (!foreground_mode && pidfile)
 		unlink(pidfile);
+#ifdef HAVE_IRQBALANCEUI
 	/* Remove socket */
 	if (socket_fd > 0)
 		close(socket_fd);
 	if (socket_name[0])
 		unlink(socket_name);
-
+#endif
 	return ret;
 }
